# BSD 3-Clause License
#
# Copyright (c) 2025, Arm Limited
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Zoomable and pannable canvas.
This will serve as the main body to which CMN objects are rendered.
"""

from PySide6.QtWidgets import QWidget, QGraphicsView, QGraphicsScene, QGraphicsSimpleTextItem
from PySide6.QtGui import QPaintEvent, QPainter, QWheelEvent, QMouseEvent, QResizeEvent, QPen, QBrush, QColor, QFont, QFontMetrics
from PySide6.QtCore import Qt, QObject, QRectF, QPointF, QSize, QEvent
from typing import cast, Callable, Tuple, Optional, Any
from functools import lru_cache

from .config import Config


class ZoomPanCanvas(QWidget):
    """
    A PySide6 QWidget subclass that supports zooming and panning.
    This canvas allows users to zoom in and out and pan across the content.

    It also supports a hover callback for mouse movement in world coordinates.
    Also a redraw callback automaticalled called when redraw required.
    Attributes:
        zoom_scale : float
            The current zoom scale factor.
        offset_x : float
            Horizontal offset for panning.
        offset_y : float
            Vertical offset for panning.
        redraw_callback : Callable[[], None]
            Callback to trigger canvas redraw; must take zero arguments.
    """
    def __init__(self, master: QWidget, redraw_callback: Callable[[], None], **kwargs: Any) -> None:
        """
        Args:
            master : QWidget
                The parent container
            redraw_callback : Callable[[], None]
                A callback function that is called whenever the canvas
                needs to be redrawn, such as after panning or zooming.
                This function must take no parameters.
            kwargs : dict
                Additional keyword arguments passed on to parent class
        """
        super().__init__(master, **kwargs)
        self.zoom_scale: float = Config.DEFAULT_ZOOM
        self.offset_x: float = 0.0
        self.offset_y: float = 0.0
        self._pan_start: Optional[QPointF] = None
        self.redraw_callback: Callable[[], None] = redraw_callback

        # Set up QGraphicsScene and QGraphicsView
        # Scene holds items to be rendered
        # View is what those items are rendered to
        self.scene: QGraphicsScene = QGraphicsScene(self)
        self.scene.setSceneRect(-500000, -500000, 1000000, 1000000)

        self.view: QGraphicsView = QGraphicsView(self.scene, self)
        self.view.setRenderHints(QPainter.RenderHint.Antialiasing)
        self.view.setTransformationAnchor(QGraphicsView.ViewportAnchor.NoAnchor)
        self.view.setResizeAnchor(QGraphicsView.ViewportAnchor.NoAnchor)
        self.view.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.view.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.view.setDragMode(QGraphicsView.DragMode.NoDrag)

        # Event handlers
        self.view.viewport().installEventFilter(self)

        # cache drawing objects
        self._brush = QBrush(QColor("black"))
        self._pen = QPen()

    def get_dynamic_grid_cell_size(self) -> float:
        """Scale Current Grid Size."""
        return Config.GRID_CELL_SIZE * (1.0 + (self.zoom_scale - 1.0) * 0.3)

    def _start_pan(self, event: QMouseEvent) -> None:
        """Start panning operation."""
        self._pan_start = event.position()

    def _do_pan(self, event: QMouseEvent) -> None:
        """Handle panning motion."""
        if not self._pan_start:
            return

        dx = event.position().x() - self._pan_start.x()
        dy = event.position().y() - self._pan_start.y()
        self.offset_x += dx
        self.offset_y += dy
        self._pan_start = event.position()
        self.redraw_callback()

    def _do_zoom(self, event: QWheelEvent) -> None:
        """Handle zoom operation from mouse wheel or keyboard shortcuts."""
        angle_delta = event.angleDelta().y()
        if angle_delta == 0:
            return
        if angle_delta > 0 and self.zoom_scale >= Config.MAX_ZOOM:
            return
        if angle_delta < 0 and self.zoom_scale <= Config.MIN_ZOOM:
            return

        factor = Config.ZOOM_FACTOR if angle_delta > 0 else 1 / Config.ZOOM_FACTOR
        mouse_x = event.position().x()
        mouse_y = event.position().y()

        new_scale = max(Config.MIN_ZOOM, min(self.zoom_scale * factor, Config.MAX_ZOOM))
        if new_scale != self.zoom_scale:
            old_grid_size = self.get_dynamic_grid_cell_size()
            old_world_x = (mouse_x - self.offset_x) / self.zoom_scale
            old_world_y = (mouse_y - self.offset_y) / self.zoom_scale
            self.zoom_scale = new_scale
            new_grid_size = self.get_dynamic_grid_cell_size()
            grid_scale_factor = new_grid_size / old_grid_size
            new_world_x = old_world_x * grid_scale_factor
            new_world_y = old_world_y * grid_scale_factor
            self.offset_x = mouse_x - new_world_x * self.zoom_scale
            self.offset_y = mouse_y - new_world_y * self.zoom_scale

            self.redraw_callback()

    def paintEvent(self, event: QPaintEvent) -> None:
        self.redraw_callback()
        return super().paintEvent(event)

    def resizeEvent(self, event: QResizeEvent):
        """Handle canvas resize."""
        super().resizeEvent(event)
        self.view.resize(QSize(event.size().width(), event.size().height()))

    def _on_mouse_move(self, event: QMouseEvent) -> None:
        """Handle mouse movement for hover effects."""
        pass

    def eventFilter(self, source: QObject, event: QEvent) -> bool:
        """Handle all mouse and keyboard events."""
        if source is self.view.viewport():
            if event.type() == event.Type.Wheel:
                self._do_zoom(cast(QWheelEvent, event))
                return True

            elif event.type() == event.Type.MouseButtonPress:
                event = cast(QMouseEvent, event)
                if event.button() == Qt.MouseButton.LeftButton:
                    self._start_pan(event)
                    return True

            elif event.type() == event.Type.MouseMove:
                if self._pan_start:
                    self._do_pan(cast(QMouseEvent, event))
                else:
                    self._on_mouse_move(cast(QMouseEvent, event))
                return True

            elif event.type() == event.Type.MouseButtonRelease:
                event = cast(QMouseEvent, event)
                if event.button() == Qt.MouseButton.LeftButton:
                    self._pan_start = None
                    return True
        return super().eventFilter(source, event)

    def world_to_screen(self, x: float, y: float) -> Tuple[float, float]:
        """Convert world coordinates to screen coordinates."""
        return x * self.zoom_scale + self.offset_x, y * self.zoom_scale + self.offset_y

    # Line thickness constants
    MIN_THICKNESS = 1
    MAX_THICKNESS = 3

    def draw_line(
            self,
            x1: float, y1: float,
            x2: float, y2: float,
            color: QColor, thickness: float = 1.0,
            scale_rate: float = 1.0,
            data: Any = None) -> None:
        """Draw a line in world coordinates."""

        sx1, sy1 = self.world_to_screen(x1, y1)
        sx2, sy2 = self.world_to_screen(x2, y2)
        effective_scale = self.zoom_scale ** scale_rate
        base_thickness = thickness * effective_scale
        scaled_thickness = max(self.MIN_THICKNESS, min(self.MAX_THICKNESS, base_thickness))

        self._pen.setColor(QColor(color))
        self._pen.setWidthF(scaled_thickness)
        item = self.scene.addLine(sx1, sy1, sx2, sy2, self._pen)
        if data is not None:
            item.setData(0, data)

    def draw_rectangle(
            self,
            x: float, y: float,
            width: float, height: float,
            color: QColor, outline_color: QColor = QColor("black"),
            outline_thickness: float = 1.0,
            scale_rate: float = 1.0,
            data: Any = None) -> None:
        """Draw a filled rectangle in world coordinates."""

        sx, sy = self.world_to_screen(x, y)
        effective_scale = self.zoom_scale ** scale_rate
        scaled_width = width * effective_scale
        scaled_height = height * effective_scale
        base_thickness = outline_thickness * effective_scale
        scaled_thickness = max(self.MIN_THICKNESS, min(self.MAX_THICKNESS, base_thickness))

        rect = QRectF(sx - scaled_width / 2, sy - scaled_height / 2, scaled_width, scaled_height)

        self._brush.setColor(color)
        self._pen.setColor(outline_color)
        self._pen.setWidthF(scaled_thickness)
        item = self.scene.addRect(rect, self._pen, self._brush)
        if data is not None:
            item.setData(0, data)

    def draw_outline_rectangle(
            self,
            x: float, y: float,
            width: float, height: float,
            outline_color: QColor, thickness: float = 1.0,
            scale_rate: float = 1.0,
            data: Any = None) -> None:
        """Draw an outline rectangle in world coordinates."""

        sx, sy = self.world_to_screen(x, y)
        effective_scale = self.zoom_scale ** scale_rate
        scaled_width = width * effective_scale
        scaled_height = height * effective_scale
        base_thickness = thickness * effective_scale
        scaled_thickness = max(self.MIN_THICKNESS, min(self.MAX_THICKNESS, base_thickness))

        rect = QRectF(sx - scaled_width / 2, sy - scaled_height / 2, scaled_width, scaled_height)
        self._pen.setColor(outline_color)
        self._pen.setWidthF(scaled_thickness)
        item = self.scene.addRect(rect, self._pen)
        if data:
            item.setData(0, data)

    def draw_text(self, x, y, text: str, color: QColor = QColor("white"), font_size=12, angle=0, scale_rate=1.0, data=None):
        """Draw text using QGraphicsSimpleTextItem - faster than addText."""
        sx, sy = self.world_to_screen(x, y)
        effective_scale = self.zoom_scale ** scale_rate
        scaled_font_size = max(1, int(font_size * effective_scale))
        
        item = QGraphicsSimpleTextItem(text)
        item.setFont(self._get_font(scaled_font_size))
        self._brush.setColor(color)
        item.setBrush(self._brush)

        self.scene.addItem(item)
        bounding = item.boundingRect()
        item.setPos(sx - bounding.width() / 2, sy - bounding.height() / 2)
        if angle != 0:
            item.setTransformOriginPoint(bounding.center())
            item.setRotation(angle)
        if data is not None:
            item.setData(0, data)

    @staticmethod
    @lru_cache(maxsize=128)
    def _get_font(scaled_size: int) -> QFont:
        """Cache Font objects"""
        font = QFont()
        font.setPointSize(scaled_size)
        return font

    @staticmethod
    @lru_cache(maxsize=4096)
    def _measure_text_width_scaled(text: str, scaled_size: int) -> float:
        font = ZoomPanCanvas._get_font(scaled_size)
        metrics = QFontMetrics(font)
        return float(metrics.horizontalAdvance(str(text)))

    @staticmethod
    @lru_cache(maxsize=128)
    def _measure_text_height_scaled(scaled_size: int) -> float:
        font = ZoomPanCanvas._get_font(scaled_size)
        metrics = QFontMetrics(font)
        return float(metrics.height())

    def get_text_width(self, text: str, font_size: int, scale_rate: float = 1.0) -> float:
        zoom_scale = self.zoom_scale
        effective_scale = zoom_scale ** scale_rate
        scaled_size = max(1, int(font_size * effective_scale))
        width_px = ZoomPanCanvas._measure_text_width_scaled(text, scaled_size)
        return width_px / effective_scale

    def get_text_height(self, font_size: int, scale_rate: float = 1.0) -> float:
        zoom_scale = self.zoom_scale
        effective_scale = zoom_scale ** scale_rate
        scaled_size = max(1, int(font_size * effective_scale))
        height_px = ZoomPanCanvas._measure_text_height_scaled(scaled_size)
        return height_px / effective_scale



# BSD 3-Clause License
#
# Copyright (c) 2025, Arm Limited
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Topology file loading and validation utilities.
"""

import json
from typing import Dict, Any


class cmn_topology_loader:
    """
    This class provides functionality to load topology data from JSONs.
    This will include loading from a file or calling wperf and capturing output.

    Attributes:
        data (Dict[str, Any]): Loaded topology data from JSON file
    """
    def __init__(self):
        self.data: Dict[str, Any] = {}

    def load_topology_from_file(self, filename: str) -> None:
        """Load topology from JSON file."""
        try:
            with open(filename, "r") as file:
                data = json.load(file)
        except FileNotFoundError:
            raise FileNotFoundError(f"Topology file '{filename}' not found")
        except json.JSONDecodeError:
            raise ValueError(f"Invalid JSON in topology file '{filename}'")

        assert data, f"Parsed topology file: {filename} is empty"
        self.data = data



# BSD 3-Clause License
#
# Copyright (c) 2025, Arm Limited
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pandas as pd
import pandas.testing as pdt
from unittest.mock import mock_open, patch
from wperf_cmn_visualizer.telemetry_loader import cmn_telemetry_loader
from wperf_cmn_visualizer.telemetry_loader import TIME_COLUMN_NAME, METRICS_COLUMN_NAME, NODE_COLUMN_NAME, VALUE_COLUMN_NAME


class TestTelemetryLoader:
    """Tests for the telemetry loader functionality."""

    def test_load_valid_csv_file(self):
        """Test loading a valid CSV telemetry file."""
        csv_content = """run,time,level,stage,group,metric,node,value,interrupted,units
1,0.500514296,2,0,,HNF SLC miss ratio with filters = hbt_lbt_filter.all_requests_selected,Global,45.3351605662493,No,Percentage
1,0.500514296,2,0,,HNF SLC miss ratio with filters = hbt_lbt_filter.all_requests_selected,HN-F (144) at X=1 Y=2 Port=0,48.247426210951204,No,Percentage"""

        expected_data = pd.DataFrame([
            {
                TIME_COLUMN_NAME: 0.500514296,
                METRICS_COLUMN_NAME: "HNF SLC miss ratio with filters = hbt_lbt_filter.all_requests_selected",
                NODE_COLUMN_NAME: "Global",
                VALUE_COLUMN_NAME: 45.3351605662493
            },
            {
                TIME_COLUMN_NAME: 0.500514296,
                METRICS_COLUMN_NAME: "HNF SLC miss ratio with filters = hbt_lbt_filter.all_requests_selected",
                NODE_COLUMN_NAME: "HN-F (144) at X=1 Y=2 Port=0",
                VALUE_COLUMN_NAME: 48.247426210951204
            }
        ])
        expected_data[METRICS_COLUMN_NAME] = expected_data[METRICS_COLUMN_NAME].astype("category")
        expected_data[NODE_COLUMN_NAME] = expected_data[NODE_COLUMN_NAME].astype("string")

        with patch("builtins.open", mock_open(read_data=csv_content)):
            loader = cmn_telemetry_loader()
            loader.load_telemetry_from_file("test.csv")
            pdt.assert_frame_equal(loader.data.reset_index(drop=True), expected_data)

    def test_load_empty_csv_file(self):
        """Test loading an empty CSV file with headers only."""
        csv_content = "run,time,level,stage,group,metric,node,value,interrupted,units\n"

        with patch("builtins.open", mock_open(read_data=csv_content)):
            loader = cmn_telemetry_loader()
            loader.load_telemetry_from_file("empty.csv")
            assert loader.data.empty

    def test_load_nonexistent_file(self):
        """Test error handling when file doesn't exist."""
        loader = cmn_telemetry_loader()

        with patch("builtins.open", side_effect=FileNotFoundError("File not found")):
            with patch("builtins.print") as mock_print:
                loader.load_telemetry_from_file("nonexistent.csv")
                mock_print.assert_called_once_with("Error: File 'nonexistent.csv' not found.")
                assert loader.data.empty

    def test_initial_state(self):
        """Test the initial state of the loader."""
        loader = cmn_telemetry_loader()
        assert isinstance(loader.data, pd.DataFrame)
        assert loader.data.empty



# BSD 3-Clause License
#
# Copyright (c) 2025, Arm Limited
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
CMN metrics data structures and operations.
"""

from PySide6.QtGui import QPalette, QColor
import numpy as np
import pandas as pd
from typing import Dict
from .cmn import CMN_MAX_PORTS, CMN_MAX_CHILDS, CMN_MAX_MESH_WIDTH, CMN_MAX_MESH_HEIGHT
from .cmn import CMN
from .telemetry_loader import TIME_COLUMN_NAME, METRICS_COLUMN_NAME, NODE_COLUMN_NAME, VALUE_COLUMN_NAME


class CMNMetrics:
    """
    Handles CMN metrics by storing them in structured numpy arrays
    indexed by time, mesh, and coordinates within the mesh.

    TODO:
        - support the use of multiple metrics at the same time.
        - support Global metrics

    NOTE:
        -CMNMetrics only supports one mesh. This is because input data from \
            Telemetry Solution's `topdown-tool` only supports 1 mesh.

    Attributes:
        self.metric_names (np.ndarray[dtype=str]): Compact list of unique names indexed by metric ids.
        self.metric_id_map (Dict[str, int]): Map from metric string to metric id
        self.time_stamps (np.ndarray[dtype=f8]): 1d numpy array of unique time stamps which which data is available.
        self.num_time_stamps (int): number of time_stamps
        self.data (np.nradday): 4d numpy array, index with self.data[time_stamp, mesh_index, XP coord Y, XP coord X]
    """
    def __init__(self, cmn: CMN, metrics_data: pd.DataFrame, palette: QPalette) -> None:
        """
        Contruct CMNMesh object.
        Args:
            cmn (CMN): Read-only CMN object used to validate input data. Invalid input is ignored.
            metrics_data (pd.DataFrame): Data frame from pandas loaded from cmn_telemtry_loader.
        """

        # Metric names and corresponding IDs (categorical for compactness)
        self.metric_names: np.ndarray = metrics_data[METRICS_COLUMN_NAME].cat.categories.to_numpy(dtype=str)
        self.metric_id_map: Dict[str, int] = {name: idx for idx, name in enumerate(self.metric_names)}
        self.num_metrics: int = len(self.metric_names)

        # Extract coordinate and node info using regex from the 'node' column
        # node column has three formats:
        # 1> "global" which indicates the entire mesh
        # 2> "XP at X=0 Y=0" which indicates a specific XP at given coordinate
        # 3> HN-F (101) at X=0 Y=0 Port=0 which indicates a device on a specific XP and port. Device identified using nodeid.
        node_info: pd.DataFrame = metrics_data[NODE_COLUMN_NAME].str.extract(
            r"^(?:(?P<device>.+?) \((?P<nodeid>\d+)\) at X=(?P<x_device>\d+) Y=(?P<y_device>\d+) Port=(?P<port>\d+)"
            r"|(?P<device_xp>XP) at X=(?P<x_xp>\d+) Y=(?P<y_xp>\d+)"
            r"|(?P<global>Global))$"
        )
        # remove outliers
        metrics_data, node_info = self._remove_outliers_mean_std(metrics_data, node_info)

        self.time_stamps: np.ndarray = np.sort(metrics_data[TIME_COLUMN_NAME].unique())
        self.num_time_stamps: int = len(self.time_stamps)

        # data holding members: Global level, XP level, Port level, Device level
        self.global_data: np.ndarray = np.zeros(
            shape=(self.num_time_stamps, self.num_metrics, cmn.num_meshes),
            dtype="f8"
        )
        self.xp_data: np.ndarray = np.zeros(
            shape=(
                self.num_time_stamps, self.num_metrics, cmn.num_meshes,
                CMN_MAX_MESH_HEIGHT, CMN_MAX_MESH_WIDTH
            ),
            dtype="f8"
        )
        self.port_data: np.ndarray = np.zeros(
            shape=(
                self.num_time_stamps, self.num_metrics, cmn.num_meshes,
                CMN_MAX_MESH_HEIGHT, CMN_MAX_MESH_WIDTH,
                CMN_MAX_PORTS
            ),
            dtype="f8"
        )
        self.device_data: np.ndarray = np.zeros(
            shape=(
                self.num_time_stamps, self.num_metrics, cmn.num_meshes,
                CMN_MAX_MESH_HEIGHT, CMN_MAX_MESH_WIDTH,
                CMN_MAX_PORTS, CMN_MAX_CHILDS
            ),
            dtype="f8"
        )

        time_indices = np.searchsorted(self.time_stamps, metrics_data[TIME_COLUMN_NAME].to_numpy())
        metric_ids = metrics_data[METRICS_COLUMN_NAME].cat.codes.to_numpy()
        metric_values = metrics_data[VALUE_COLUMN_NAME].to_numpy()

        self._load_global_data(node_info, time_indices, metric_ids, metric_values)
        self._load_device_data(node_info, cmn, time_indices, metric_ids, metric_values)
        self._load_xp_data(node_info, time_indices, metric_ids, metric_values)

        # cache for get_metric_min_max method
        self._last_metric_id = None
        self._last_mesh_idx = None
        self._last_min_max = (0.0, 0.0)
        self.palette: QPalette = palette
        self.base_color = self.palette.color(QPalette.ColorRole.Window)

    def _remove_outliers_mean_std(
        self,
        df: pd.DataFrame, node_info: pd.DataFrame,
        std_multiplier: float = 2.0
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        """
        Properly removes outliers by filtering per-metric and per-node-type.
        Keeps values within [mean - N*std, mean + N*std] AND < 1e6.
        """
        keep_mask = pd.Series(False, index=df.index)

        global_mask = node_info["global"].notna()
        non_global_mask = ~global_mask

        for metric in self.metric_names:
            metric_mask = df[METRICS_COLUMN_NAME] == metric

            # GLOBAL
            subset_global = df[metric_mask & global_mask]
            subset_global = subset_global[subset_global[VALUE_COLUMN_NAME] < 1e10]

            if len(subset_global) >= 4:
                values = subset_global[VALUE_COLUMN_NAME]
                mean = values.mean()
                std = values.std()
                lower = mean - std_multiplier * std
                upper = mean + std_multiplier * std
                valid_mask = (values >= lower) & (values <= upper)
                keep_mask.loc[subset_global[valid_mask].index] = True
            else:
                keep_mask.loc[subset_global.index] = True

            # NON-GLOBAL
            subset_local = df[metric_mask & non_global_mask]
            subset_local = subset_local[subset_local[VALUE_COLUMN_NAME] < 1e10]

            if len(subset_local) >= 4:
                values = subset_local[VALUE_COLUMN_NAME]
                mean = values.mean()
                std = values.std()
                lower = mean - std_multiplier * std
                upper = mean + std_multiplier * std
                valid_mask = (values >= lower) & (values <= upper)
                keep_mask.loc[subset_local[valid_mask].index] = True
            else:
                keep_mask.loc[subset_local.index] = True

        filtered_df = df[keep_mask].reset_index(drop=True)
        filtered_node_info = node_info[keep_mask].reset_index(drop=True)

        return filtered_df, filtered_node_info

    def _load_global_data(self, node_info, time_indices, metric_ids, metric_values) -> None:
        global_mask = node_info["global"].notna()
        if global_mask.any():
            global_rows = np.where(global_mask)[0]
            global_t = time_indices[global_rows]
            global_m = np.zeros(len(global_rows), dtype=int)  # mesh = 0
            global_ids = metric_ids[global_rows]
            global_vals = metric_values[global_rows]
            self.global_data[global_t, global_ids, global_m] = global_vals

    def _load_xp_data(self, node_info, time_indices, metric_ids, metric_values) -> None:
        x_xp = node_info["x_xp"].astype("float")
        y_xp = node_info["y_xp"].astype("float")
        xp_mask = (
            x_xp.notna() & y_xp.notna()
            & (x_xp >= 0) & (x_xp < CMN_MAX_MESH_WIDTH)
            & (y_xp >= 0) & (y_xp < CMN_MAX_MESH_HEIGHT)
        )

        if xp_mask.any():
            xp_rows = np.where(xp_mask)[0]
            xp_t = time_indices[xp_rows]
            xp_m = np.zeros(len(xp_rows), dtype=int)
            xp_y = node_info.loc[xp_mask, "y_xp"].astype(int).to_numpy()
            xp_x = node_info.loc[xp_mask, "x_xp"].astype(int).to_numpy()
            xp_ids = metric_ids[xp_rows]
            xp_vals = metric_values[xp_rows]
            self.xp_data[xp_t, xp_ids, xp_m, xp_y, xp_x] = xp_vals

    def _load_device_data(self, node_info, cmn, time_indices, metric_ids, metric_values) -> None:
        x_dev = node_info["x_device"].astype("float")
        y_dev = node_info["y_device"].astype("float")
        port_dev = node_info["port"].astype("float")

        device_mask = (
            x_dev.notna() & y_dev.notna() & port_dev.notna()
            & (x_dev >= 0) & (x_dev < CMN_MAX_MESH_WIDTH)
            & (y_dev >= 0) & (y_dev < CMN_MAX_MESH_HEIGHT)
            & (port_dev >= 0) & (port_dev < CMN_MAX_PORTS)
        )

        if device_mask.any():
            dev_rows = np.where(device_mask)[0]
            dev_t = time_indices[dev_rows]
            dev_nodeids = node_info.loc[device_mask, "nodeid"].astype(int)
            dev_y = node_info.loc[device_mask, "y_device"].astype(int)
            dev_x = node_info.loc[device_mask, "x_device"].astype(int)
            dev_ports = node_info.loc[device_mask, "port"].astype(int)
            dev_ids = metric_ids[dev_rows]
            dev_vals = metric_values[dev_rows]

            for t, nodeid, y, x, port, mid, val in zip(
                dev_t, dev_nodeids, dev_y, dev_x, dev_ports, dev_ids, dev_vals
            ):
                for device_idx, device in enumerate(cmn.meshes[0]["xps"][y, x]["ports"][port]["devices"]):
                    if device["nodeid"] == nodeid:
                        self.device_data[t, mid, 0, y, x, port, device_idx] = val

            # aggregate metric values for each device onto its port
            non_zero_dev_mask = self.device_data != 0.0
            device_counts = np.count_nonzero(non_zero_dev_mask, axis=6, keepdims=False)
            device_sums = self.device_data.sum(axis=6)
            with np.errstate(divide='ignore', invalid='ignore'):
                port_means = np.where(
                    device_counts > 0,
                    device_sums / device_counts,
                    0.0
                )
            self.port_data = port_means

            # aggregate metrics values for each port onto its XP
            non_zero_port_mask = self.port_data != 0.0
            port_counts = np.count_nonzero(non_zero_port_mask, axis=5, keepdims=False)
            port_sums = self.port_data.sum(axis=5)
            with np.errstate(divide='ignore', invalid='ignore'):
                xp_means = np.where(
                    port_counts > 0,
                    port_sums / port_counts,
                    0.0
                )
            self.xp_data = xp_means

    def get_metric_min_max(self, metric_id: int, mesh_idx: int) -> tuple[float, float]:
        """
        Compute the minimum and maximum non-zero values for a given metric and mesh.
        Args:
            metric_id: Metric identifier to filter by.
            mesh_idx: Index of the mesh to extract from.
        Returns:
            A tuple containing (min_value, max_value) for the selected metric.
            If no non-zero values are found, returns (0.0, 0.0).
        """
        if metric_id == self._last_metric_id and mesh_idx == self._last_mesh_idx:
            return self._last_min_max

        metric_id = max(0, min(metric_id, self.num_metrics - 1))
        mesh_idx = max(0, min(mesh_idx, self.global_data.shape[2] - 1))

        min_val = np.inf
        max_val = -np.inf

        # XP level
        xp_vals = self.xp_data[:, metric_id, mesh_idx, :, :]
        xp_mask = xp_vals != 0.0
        if np.any(xp_mask):
            xp_filtered = xp_vals[xp_mask]
            min_val = min(min_val, xp_filtered.min())
            max_val = max(max_val, xp_filtered.max())

        # Port level
        port_vals = self.port_data[:, metric_id, mesh_idx, :, :, :]
        port_mask = port_vals != 0.0
        if np.any(port_mask):
            port_filtered = port_vals[port_mask]
            min_val = min(min_val, port_filtered.min())
            max_val = max(max_val, port_filtered.max())

        # Device level
        dev_vals = self.device_data[:, metric_id, mesh_idx, :, :, :, :]
        dev_mask = dev_vals != 0.0
        if np.any(dev_mask):
            dev_filtered = dev_vals[dev_mask]
            min_val = min(min_val, dev_filtered.min())
            max_val = max(max_val, dev_filtered.max())

        if min_val == np.inf:
            result = (0.0, 0.0)
        else:
            result = (float(min_val), float(max_val))

        self._last_metric_id = metric_id
        self._last_mesh_idx = mesh_idx
        self._last_min_max = result

        return result

    def _value_to_colour(self, value: float, vmin: float, vmax: float) -> QColor:
        """Blend between window colour and bright red based on value."""
        if vmax <= vmin:
            return self.base_color

        norm = (value - vmin) / (vmax - vmin)
        norm = max(0.0, min(norm, 1.0))
        inv_norm = 1 - norm
        r = int(self.base_color.red() * inv_norm + 255.0 * norm)
        g = int(self.base_color.green() * inv_norm)
        b = int(self.base_color.blue() * inv_norm)
        return QColor(r, g, b)

    def get_xp_colour(self, time_idx: int, metric_id: int, mesh_idx: int, y: int, x: int) -> QColor:
        val = self.xp_data[time_idx, metric_id, mesh_idx, y, x]
        vmin, vmax = self.get_metric_min_max(metric_id, mesh_idx)
        return self._value_to_colour(val, vmin, vmax)

    def get_port_colour(self, time_idx: int, metric_id: int, mesh_idx: int, y: int, x: int, port: int) -> QColor:
        val = self.port_data[time_idx, metric_id, mesh_idx, y, x, port]
        vmin, vmax = self.get_metric_min_max(metric_id, mesh_idx)
        return self._value_to_colour(val, vmin, vmax)

    def get_device_colour(self, time_idx: int, metric_id: int, mesh_idx: int, y: int, x: int, port: int, device_idx: int) -> QColor:
        val = self.device_data[time_idx, metric_id, mesh_idx, y, x, port, device_idx]
        vmin, vmax = self.get_metric_min_max(metric_id, mesh_idx)
        return self._value_to_colour(val, vmin, vmax)



# BSD 3-Clause License
#
# Copyright (c) 2025, Arm Limited
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Time Line Scrubbing Module.

Instantiate a TimeScrubber widget which broadcasts a time scrubbing event.
Show Global data as a time line graph and present media buttons.
"""

from PySide6.QtWidgets import (
    QWidget, QHBoxLayout, QComboBox, QToolButton, QFrame
)
from PySide6.QtCore import Qt, QTimer, Signal, QPointF
from PySide6.QtGui import QPainter, QPen, QColor, QPaintEvent, QMouseEvent, QIcon, QPolygonF
import numpy as np
from typing import Any

from .cmn_metrics import CMNMetrics


class TimelineCanvas(QWidget):
    """Widget for drawing the timeline graph and handling scrubbing."""

    scrub_changed = Signal(int)
    """Signal for Scrubbing. Emitted when time stamp index is changed"""

    def __init__(self, master: QWidget, cmn_metrics: CMNMetrics, **kwargs):
        """
        Initialize the status bar.
        Args:
            master: Parent widget
            cmn_metrics: Metrics data source
            **kwargs: Additional widget configuration options
        """
        super().__init__(master, **kwargs)
        self.cmn_metrics = cmn_metrics
        self.current_time_index = 0
        self.handle_x = 0.0
        self.is_dragging = False
        self.is_hovering = False

        self.setMouseTracking(True)
        self.setCursor(Qt.CursorShape.ArrowCursor)

    def paintEvent(self, event: QPaintEvent) -> None:
        """Paint the graph line and scrubbing handle."""
        painter = QPainter(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)
        width = self.width()
        height = self.height()
        if width > 1 and height > 1:
            self._draw_graph_line(painter, width, height)
            self._draw_scrubbing_handle(painter, width, height)

    def _draw_graph_line(self, painter: QPainter, width: int, height: int) -> None:
        """
        Draw the metric graph line normalised to canvas size.
        Args:
            width: Canvas width.
            height: Canvas height.
        """
        if self.cmn_metrics.global_data.size == 0:
            return

        values = self.cmn_metrics.global_data[:, 0, 0]  # Assuming metric_idx=0, mesh=0
        if len(values) == 0:
            return

        max_val = np.max(values)
        if max_val == 0:
            return

        normalized_y = values / max_val
        y_coords = (1.0 - normalized_y) * (height - 5)

        max_time = np.max(self.cmn_metrics.time_stamps) if len(self.cmn_metrics.time_stamps) > 0 else 1
        normalized_x = self.cmn_metrics.time_stamps / max_time
        x_coords = normalized_x * (width - 5)

        pen = QPen(QColor(70, 130, 180), 2)
        painter.setPen(pen)
        polygon = QPolygonF([QPointF(x, y) for x, y in zip(x_coords, y_coords)])
        painter.setPen(QPen(QColor(70, 130, 180), 2))
        painter.drawPolyline(polygon)

    def _draw_scrubbing_handle(self, painter: QPainter, width: int, height: int) -> None:
        """
        Draw the scrubbing handle at the current time index.
        Args:
            width: Canvas width.
            height: Canvas height.
            thick: If True, draw thicker and semi-transparent handle.
        """
        if len(self.cmn_metrics.time_stamps) == 0:
            return

        max_time = np.max(self.cmn_metrics.time_stamps)
        current_time = self.cmn_metrics.time_stamps[self.current_time_index]
        self.handle_x = (current_time / max_time) * (width - 5)

        if self.is_dragging or self.is_hovering:
            colour = QColor(220, 20, 60, 100)
            pen = QPen(colour, 8)
        else:
            colour = QColor(220, 20, 60)
            pen = QPen(colour, 4)
        pen.setCapStyle(Qt.PenCapStyle.RoundCap)
        painter.setPen(pen)
        painter.drawLine(int(self.handle_x), 0 + 5, int(self.handle_x), height - 5)

    def _get_nearest_time_index(self, x: float) -> int:
        """
        Find the nearest time index for a given x coordinate.
        Args:
            x: X coordinate on the canvas.
        Returns:
            Closest index in time_stamps.
        """
        if len(self.cmn_metrics.time_stamps) == 0:
            return 0

        width = self.width()
        max_time = np.max(self.cmn_metrics.time_stamps)
        normalized_time = x / (width - 5)
        target_time = normalized_time * max_time

        distances = np.abs(self.cmn_metrics.time_stamps - target_time)
        return int(np.argmin(distances))

    def _is_near_handle(self, x: float, tolerance: int = 10) -> bool:
        """
        Check if mouse is near the scrubbing handle.
        Args:
            x: Mouse x coordinate.
            tolerance: Pixel tolerance for proximity.
        Returns:
            True if mouse is near the handle, else False.
        """
        return abs(x - self.handle_x) <= tolerance

    def mousePressEvent(self, event: QMouseEvent) -> None:
        """Start dragging if mouse press near the handle."""
        if event.button() == Qt.MouseButton.LeftButton and self._is_near_handle(event.position().x()):
            self.is_dragging = True

    def mouseMoveEvent(self, event: QMouseEvent) -> None:
        """Update scrubbing position during dragging or handle hover effects."""
        x = event.position().x()
        if self.is_dragging:
            x = max(5, min(x, self.width() - 5))
            new_index = self._get_nearest_time_index(x)
            if new_index != self.current_time_index:
                self.current_time_index = new_index
                self.update()  # Trigger repaint
                self.scrub_changed.emit(self.current_time_index)
        else:
            was_hovering = self.is_hovering
            self.is_hovering = self._is_near_handle(x)
            if self.is_hovering != was_hovering:
                self.setCursor(Qt.CursorShape.SizeHorCursor if self.is_hovering else Qt.CursorShape.ArrowCursor)
                self.update()

    def mouseReleaseEvent(self, event: QMouseEvent) -> None:
        """Stop dragging."""
        if event.button() == Qt.MouseButton.LeftButton:
            self.is_dragging = False
            self.update()

    def leaveEvent(self, event) -> None:
        """Reset hover state when mouse leaves widget."""
        if not self.is_dragging:
            self.is_hovering = False
            self.setCursor(Qt.CursorShape.ArrowCursor)
            self.update()

    def set_time_index(self, index: int) -> None:
        """Set the current time index and update display."""
        if 0 <= index < len(self.cmn_metrics.time_stamps):
            self.current_time_index = index
            self.update()


class TimeScrubber(QWidget):
    """Status bar widget with media playback controls and custom render area."""

    # Signal emitted when time position changes
    time_changed = Signal(int)

    def __init__(self, master: QWidget, cmn_metrics: CMNMetrics, height: int, **kwargs: Any) -> None:
        """
        Initialize the time scrubber.
        Args:
            master: Parent widget
            cmn_metrics: Metrics data source
            **kwargs: Additional configuration options
        """
        super().__init__(master)
        self.master = master
        self.cmn_metrics = cmn_metrics
        self.desired_height = height

        # Playback state
        self.is_playing = False
        self.playback_speed = 1.0
        self.current_time_index = 0

        # Timer for playback
        self.playback_timer = QTimer()
        self.playback_timer.timeout.connect(self._play_next_frame)

        self._setup_ui()

    def _setup_ui(self) -> None:
        """Set up the user interface."""
        self.setFixedHeight(self.desired_height)
        main_layout = QHBoxLayout(self)
        main_layout.setContentsMargins(5, 2, 5, 2)
        self._create_media_controls(main_layout)

        separator = QFrame()
        separator.setFrameShape(QFrame.Shape.VLine)
        separator.setFrameShadow(QFrame.Shadow.Sunken)
        separator.setLineWidth(1)
        main_layout.addWidget(separator)

        self.timeline_canvas = TimelineCanvas(self.master, self.cmn_metrics)
        self.timeline_canvas.scrub_changed.connect(self._on_scrub_changed)
        main_layout.addWidget(self.timeline_canvas, 1)

    def _create_media_controls(self, parent_layout: QHBoxLayout) -> None:
        """Create playback control buttons."""

        def create_button(icon_name: str, tooltip: str, callback):
            btn = QToolButton()
            btn.setIcon(QIcon.fromTheme(icon_name))
            btn.setToolTip(tooltip)
            btn.setAutoRaise(True)
            btn.setFixedSize(30, 30)
            btn.clicked.connect(callback)
            parent_layout.addWidget(btn)
            return btn

        self.play_button = create_button("media-playback-start", "Play/Pause", self._on_play_pause)
        self.stop_button = create_button("media-playback-stop", "Stop", self._on_stop)
        self.prev_button = create_button("media-skip-backward", "Previous", self._on_previous)
        self.next_button = create_button("media-skip-forward", "Next", self._on_next)

        # Speed combobox remains unchanged
        self.speed_combobox = QComboBox()
        self.speed_values = ["0.25x", "0.5x", "1x", "1.5x", "2x", "4x"]
        self.speed_combobox.addItems(self.speed_values)
        self.speed_combobox.setCurrentText("1x")
        self.speed_combobox.setFixedHeight(30)
        self.speed_combobox.currentTextChanged.connect(self._on_speed_change)
        parent_layout.addWidget(self.speed_combobox)

    def _on_play_pause(self) -> None:
        """Toggle playback state and update play button icon."""
        self.is_playing = not self.is_playing
        icon_name = "media-playback-pause" if self.is_playing else "media-playback-start"
        self.play_button.setIcon(QIcon.fromTheme(icon_name))
        if self.is_playing:
            self._start_playback()
        else:
            self.playback_timer.stop()

    def _start_playback(self) -> None:
        """Start playback from current position."""
        if self.current_time_index < len(self.cmn_metrics.time_stamps) - 1:
            self._schedule_next_frame()
        else:
            self.is_playing = False
            self.play_button.setIcon(QIcon.fromTheme("media-playback-start"))

    def _schedule_next_frame(self) -> None:
        """Schedule the next frame based on time delta."""
        if not self.is_playing or self.current_time_index >= len(self.cmn_metrics.time_stamps) - 1:
            return
        t_current = self.cmn_metrics.time_stamps[self.current_time_index]
        t_next = self.cmn_metrics.time_stamps[self.current_time_index + 1]
        delay = max(1, int(((t_next - t_current) * 1000) / self.playback_speed))
        self.playback_timer.start(delay)

    def _play_next_frame(self) -> None:
        """Advance playback by one frame."""
        self.playback_timer.stop()
        if not self.is_playing:
            return
        if self.current_time_index < len(self.cmn_metrics.time_stamps) - 1:
            self.current_time_index += 1
            self.timeline_canvas.set_time_index(self.current_time_index)
            self._broadcast_scrub_event()
            self._schedule_next_frame()
        else:
            self.is_playing = False
            self.play_button.setIcon(QIcon.fromTheme("media-playback-start"))

    def _on_stop(self) -> None:
        """Stop playback, reset time index and update UI."""
        self.is_playing = False
        self.playback_timer.stop()
        self.play_button.setIcon(QIcon.fromTheme("media-playback-start"))
        self.current_time_index = 0
        self.timeline_canvas.set_time_index(self.current_time_index)
        self._broadcast_scrub_event()

    def _on_previous(self) -> None:
        """Go to previous time index if possible."""
        if self.current_time_index > 0:
            self.current_time_index -= 1
            self.timeline_canvas.set_time_index(self.current_time_index)
            self._broadcast_scrub_event()

    def _on_next(self) -> None:
        """Go to next time index if possible."""
        if self.current_time_index < len(self.cmn_metrics.time_stamps) - 1:
            self.current_time_index += 1
            self.timeline_canvas.set_time_index(self.current_time_index)
            self._broadcast_scrub_event()

    def _on_speed_change(self, speed_text: str) -> None:
        """Update playback speed based on combobox selection."""
        try:
            self.playback_speed = float(speed_text.replace("x", ""))
        except ValueError:
            self.playback_speed = 1.0  # fallback

    def _on_scrub_changed(self, time_index: int) -> None:
        """Handle scrubbing from the timeline canvas."""
        self.current_time_index = time_index
        self._broadcast_scrub_event()

    def _broadcast_scrub_event(self) -> None:
        """Broadcast time scrub event."""
        self.time_changed.emit(self.current_time_index)



# BSD 3-Clause License
#
# Copyright (c) 2025, Arm Limited
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from unittest.mock import Mock
from PySide6.QtWidgets import QWidget, QApplication
from PySide6.QtCore import Qt, QPointF, QPoint, QEvent, QSize
from PySide6.QtGui import QWheelEvent, QMouseEvent
from wperf_cmn_visualizer.canvas import ZoomPanCanvas


class TestZoomPanCanvas:
    """Tests for ZoomPanCanvas PySide6 version"""

    @classmethod
    def setup_class(cls):
        cls.app = QApplication.instance() or QApplication([])

    def setup_method(self):
        """Create widget and mocks for each test"""
        self.parent = QWidget()
        self.redraw_mock = Mock()
        self.canvas = ZoomPanCanvas(self.parent, redraw_callback=self.redraw_mock)
        self.parent.show()

    def teardown_method(self):
        """Clean up widget"""
        self.canvas.deleteLater()
        self.parent.deleteLater()

    def test_initial_state(self):
        """Ensure proper initial values and objects"""
        assert self.canvas.zoom_scale == 1.0
        assert self.canvas.offset_x == 0.0
        assert self.canvas.offset_y == 0.0
        assert self.canvas.scene is not None
        assert self.canvas.view is not None

    def test_pan_updates_offsets(self):
        """
        Test to ensure panning works.
        Ensure redraw is called.
        """
        press_event = QMouseEvent(QEvent.Type.MouseButtonPress, QPointF(10, 10), QPointF(10, 10), Qt.MouseButton.LeftButton, Qt.MouseButton.LeftButton, Qt.KeyboardModifier.NoModifier)
        self.canvas.eventFilter(self.canvas.view.viewport(), press_event)
        assert self.canvas._pan_start == QPointF(10, 10)

        move_event = QMouseEvent(QEvent.Type.MouseMove, QPointF(20, 30), QPointF(20, 30), Qt.MouseButton.NoButton, Qt.MouseButton.LeftButton, Qt.KeyboardModifier.NoModifier)
        self.canvas.eventFilter(self.canvas.view.viewport(), move_event)

        assert self.canvas.offset_x == 10  # 20 - 10
        assert self.canvas.offset_y == 20  # 30 - 10
        self.redraw_mock.assert_called()

    def test_pan_stops_on_mouse_release(self):
        """
        Panning stops when left mouse button released
        """
        press_event = QMouseEvent(QEvent.Type.MouseButtonPress, QPointF(10, 10), QPointF(10, 10), Qt.MouseButton.LeftButton, Qt.MouseButton.LeftButton, Qt.KeyboardModifier.NoModifier)
        self.canvas.eventFilter(self.canvas.view.viewport(), press_event)
        release_event = QMouseEvent(QEvent.Type.MouseButtonRelease, QPointF(20, 30), QPointF(20, 30), Qt.MouseButton.LeftButton, Qt.MouseButton.NoButton, Qt.KeyboardModifier.NoModifier)
        self.canvas.eventFilter(self.canvas.view.viewport(), release_event)
        assert self.canvas._pan_start is None

    def test_world_to_screen_conversion(self):
        """
        Check world_to_screen coordinate conversion respects zoom and offset
        """
        self.canvas.zoom_scale = 2.0
        self.canvas.offset_x = 5.0
        self.canvas.offset_y = 10.0
        x_screen, y_screen = self.canvas.world_to_screen(3.0, 4.0)
        assert x_screen == 3.0 * 2.0 + 5.0
        assert y_screen == 4.0 * 2.0 + 10.0

    def test_zoom_in_updates_scale(self):
        """
        Simulate scrolling with Mouse Wheel.
        Check if canvas zoom scale is changed.
        Ensure redraw is called.
        """
        old_scale = self.canvas.zoom_scale
        pos = QPointF(50, 50)
        wheel_event = QWheelEvent(
            pos, pos, QPoint(0, 0), QPoint(0, 120),
            Qt.MouseButton.NoButton, Qt.KeyboardModifier.NoModifier,
            Qt.ScrollPhase.ScrollUpdate, False
        )
        self.canvas.eventFilter(self.canvas.view.viewport(), wheel_event)
        assert self.canvas.zoom_scale > old_scale
        self.redraw_mock.assert_called()

    def test_zoom_out_updates_scale(self):
        """
        Simulate scrolling with Mouse Wheel.
        Check if canvas zoom scale is changed.
        Ensure redraw is called.
        """
        self.canvas.zoom_scale = 2.0
        old_scale = self.canvas.zoom_scale
        pos = QPointF(50, 50)
        wheel_event = QWheelEvent(
            pos, pos, QPoint(0, 0), QPoint(0, -120),
            Qt.MouseButton.NoButton, Qt.KeyboardModifier.NoModifier,
            Qt.ScrollPhase.ScrollUpdate, False
        )
        self.canvas.eventFilter(self.canvas.view.viewport(), wheel_event)
        assert self.canvas.zoom_scale < old_scale
        self.redraw_mock.assert_called()

    def _create_wheel_event(self, angle_delta=120, pos=QPointF(50, 50)):
        return QWheelEvent(
            pos,
            pos,
            QPoint(0, 0),
            QPoint(0, angle_delta),
            Qt.MouseButton.NoButton,
            Qt.KeyboardModifier.NoModifier,
            Qt.ScrollPhase.ScrollUpdate,
            False
        )

    def test_zoom_scale_clamped_at_min(self):
        """Ensure zoom scale doesn't go below minimum limit."""
        self.canvas.zoom_scale = 0.1
        event = self._create_wheel_event(angle_delta=-120)
        self.canvas.wheelEvent(event)
        assert self.canvas.zoom_scale >= 0.1

    def test_zoom_scale_clamped_at_max(self):
        """Ensure zoom scale doesn't exceed maximum limit."""
        self.canvas.zoom_scale = 10.0
        event = self._create_wheel_event(angle_delta=120)
        self.canvas.wheelEvent(event)
        assert self.canvas.zoom_scale <= 10.0



# BSD 3-Clause License
#
# Copyright (c) 2025, Arm Limited
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Main application class for wperf_cmn_visualizer
"""

import argparse
from typing import Optional, List
from sys import argv
from PySide6.QtWidgets import (
    QApplication, QMainWindow, QWidget, QVBoxLayout, QSizePolicy, QFrame
)
from PySide6.QtGui import QGuiApplication
from PySide6.QtCore import QRect

from .config import Config
from .topology_loader import cmn_topology_loader
from .telemetry_loader import cmn_telemetry_loader
from .cmn import CMN
from .cmn_metrics import CMNMetrics
from .renderer import CMNRenderer
from .time_scrubber import TimeScrubber
from .tabbing import TabbedInterface


class wperfCmnVisualizer:
    """
    GUI application class for the WindowsPerf CMN Visualizer.
    Handles window creation, layout, and launching the Qt execution loop.
    """

    def __init__(self, args: argparse.Namespace):
        """
        Initialise the application with the provided topology file.
        Args:
            args (argparse.Namespace): Parsed command line arguments.
        """

        # open main window
        self.args = args
        self.app = QApplication(argv)
        self.window = QMainWindow()
        self._set_up_window()

        # load topology
        self.topology = cmn_topology_loader()
        if self.args.topology is not None:
            self.topology.load_topology_from_file(self.args.topology)

        # construct CMN object
        self.cmn: CMN = CMN(self.topology.data)

        # load telemetry if available
        self.telemetry = cmn_telemetry_loader()
        if self.args.telemetry:
            self.telemetry.load_telemetry_from_file(self.args.telemetry)
            self.cmn_metrics: Optional[CMNMetrics] = CMNMetrics(self.cmn, self.telemetry.data, self.app.palette())
        else:
            self.cmn_metrics = None

        # set up UI in main window.
        self._setup_widgets()

    def _setup_widgets(self):
        """
        Set up the main widgets and layout.
        +--------------+
        |     CMN      |
        |              |
        |______________|
        |   Time Line  |
        +--------------+
        """
        # Create central widget and main layout
        central_widget = QWidget()
        self.window.setCentralWidget(central_widget)
        main_layout = QVBoxLayout(central_widget)
        main_layout.setContentsMargins(0, 0, 0, 0)
        main_layout.setSpacing(0)

        # create tabs
        self.tabs = TabbedInterface()
        self.tabs.setContentsMargins(0, 0, 0, 0)
        main_layout.addWidget(self.tabs, 1)

        self.renderers: List[CMNRenderer] = []
        for i in range(self.cmn.num_meshes):
            self.renderers.append(
                CMNRenderer(self.tabs, self.app.palette(), self.cmn, self.cmn_metrics)
            )
            self.renderers[i].cmn_idx = i
            self.renderers[i].canvas.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
            self.tabs.add_tab(self.renderers[i].canvas, f"CMN {i}")

        # construct TimeScrubber class only if metrics are available
        if self.cmn_metrics is not None:
            separator = QFrame()
            separator.setFrameShape(QFrame.Shape.HLine)
            separator.setFrameShadow(QFrame.Shadow.Sunken)
            main_layout.addWidget(separator)

            self.time_scrubber: TimeScrubber = TimeScrubber(
                central_widget, self.cmn_metrics, height=100
            )
            self.time_scrubber.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
            main_layout.addWidget(self.time_scrubber, 0)  # no stretch, fixed height
            # connect scrubbing event broadcasted from TimeScrubber
            for renderer in self.renderers:
                self.time_scrubber.time_changed.connect(renderer._on_time_changed)

        # Update Layout of UI
        central_widget.updateGeometry()

    def _set_up_window(self):
        """
        Configure the main application window's appearance and geometry.
        """
        self.window.setWindowTitle(Config.MAIN_WINDOW_TITLE)

        # centre window on screen
        screen = QGuiApplication.primaryScreen()
        screen_size = screen.size()
        screen_width = screen_size.width()
        screen_height = screen_size.height()

        width = int(screen_width * Config.MAIN_WINDOW_INIT_SIZE_RATIO)
        height = int(screen_height * Config.MAIN_WINDOW_INIT_SIZE_RATIO)
        x = (screen_width - width) // 2
        y = (screen_height - height) // 2

        min_width, min_height = Config.MAIN_WINDOW_MIN_SIZE
        if width < min_width or height < min_height:
            width = min_width
            height = min_height

        self.window.setGeometry(QRect(x, y, width, height))
        self.window.setMinimumSize(min_width, min_height)

    def run(self) -> int:
        """
        Start the application's main event loop.
        Returns:
            int: to be used as exit code
        """
        try:
            self.window.show()
            return self.app.exec()
        except Exception as e:
            print(f"Unexpected error: {e}")
            return 1



# BSD 3-Clause License
#
# Copyright (c) 2025, Arm Limited
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Entry point for running cmn-visualizer as a module.
"""
import sys
import argparse
from . import __version__
from .config import Config
from .wperf_cmn_visualizer import wperfCmnVisualizer


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="A GUI-based visualizer for Arm CMN (Coherent Mesh Network)"
                    " systems using WindowsPerf backend"
    )
    parser.add_argument(
        "--version", action="version", version=f"{__version__}"
    )
    parser.add_argument(
        "--topology",
        type=str,
        default=Config.DEFAULT_TOPOLOGY_FILE,
        help=f"Path to the topology JSON file (default: {Config.DEFAULT_TOPOLOGY_FILE})",
    )
    parser.add_argument(
        "--telemetry",
        type=str,
        help="Path to the telemtry CSV file",
    )
    return parser.parse_args()


def main() -> int:
    args: argparse.Namespace = parse_args()
    app: wperfCmnVisualizer = wperfCmnVisualizer(args)
    return app.run()


if __name__ == "__main__":
    sys.exit(main())



# BSD 3-Clause License
#
# Copyright (c) 2025, Arm Limited
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
CMN mesh data structures and operations.
"""

import numpy as np
from numpy import dtype as np_dtype
from typing import Dict, Any, List

CMN_MAX_PORTS: int = 6
CMN_MAX_CHILDS: int = 32
CMN_MAX_MESH_WIDTH: int = 12
CMN_MAX_MESH_HEIGHT: int = 12
CMN_TYPE_STR_MAX_LEN: int = 128
CMN_XP_DEVICE_ID: int = 6

coord_dtype = np_dtype(
    [
        ("x", "u1"),
        ("y", "u1"),
    ]
)

node_dtype = np_dtype(
    [
        ("coord", coord_dtype),
        ("nodeid", "u2"),
        ("logicalid", "u2"),
        ("type", "u2"),
        ("type_str", "U128"),
    ]
)

port_dtype = np_dtype(
    [
        ("type", "u1"),
        ("type_str", "U128"),
        ("cal", "b1"),
        ("num_devices", "u1"),
        ("devices", node_dtype, CMN_MAX_CHILDS),
    ]
)

xp_dtype = np_dtype(
    [
        ("node_info", node_dtype),
        ("dtc_domain", "u1"),
        ("num_device_ports", "u1"),
        ("ports", port_dtype, CMN_MAX_PORTS),
    ]
)

mesh_dtype = np_dtype(
    [
        ("x_dim", "u1"),
        ("y_dim", "u1"),
        ("xps", xp_dtype, (CMN_MAX_MESH_HEIGHT, CMN_MAX_MESH_WIDTH)),
    ]
)


class CMN:
    """
    Top-level CMN topology container managing multiple meshes.
    Attributes:
        num_meshes (int): Number of meshes in the topology
        meshes (np.ndarray): Array of mesh structures.
    """
    def __init__(self, topology: Dict[str, Any]) -> None:
        """
        Initialise CMN topology from JSON configuration.
        Args:
            topology (Dict[str, Any]): JSON configuration
        Raises:
            AssertionError: On invalid input
        """
        assert "elements" in topology, "topology data does not contain elements information"
        self.num_meshes: int = len(topology["elements"])
        assert self.num_meshes > 0, f"topology data contained {self.num_meshes} meshes"
        self.meshes = np.zeros(self.num_meshes, dtype=mesh_dtype)

        for i, element in enumerate(topology["elements"]):
            assert "config" in element, f"element idx: {i} does not contain configuration data"
            self._init_mesh(self.meshes[i], element["config"])

    def _init_mesh(self, mesh: np.ndarray, config: Dict[str, Any]) -> None:
        """
        Initialise a single mesh element.
        Args:
            mesh (np.ndarray): The structured array element to populate
            config (Dict[str, Any]): Configuration data for the mesh
        """
        x_val = config.get("X", 0)
        y_val = config.get("Y", 0)

        x_dim = int(x_val)
        y_dim = int(y_val)

        assert 0 < x_dim <= CMN_MAX_MESH_WIDTH, f"invalid x dimension on mesh, got: {x_dim}"
        assert 0 < y_dim <= CMN_MAX_MESH_HEIGHT, f"invalid y dimension on mesh, got: {y_dim}"

        mesh["x_dim"] = x_dim
        mesh["y_dim"] = y_dim

        assert "xps" in config, "no XP information on mesh"
        assert len(config["xps"]) > 0, "empty XP information on mesh"

        self._load_data(mesh, config["xps"], x_dim, y_dim)

    def _load_data(
        self,
        mesh: np.ndarray,
        xps_data: List[Dict[str, Any]],
        x_dim: int, y_dim: int
    ) -> None:
        """
        Load XP data into the mesh structure.
        Args:
            mesh (np.ndarray): Mesh structure
            xps_data (List[Dict[str, Any]]): List of XP data
            x_dim (int): Width of mesh
            y_dim (int): Height of mesh
        """
        for xp in xps_data:
            x = int(xp.get("X", 0))
            y = int(xp.get("Y", 0))

            if not (0 <= x < x_dim and 0 <= y < y_dim):
                continue

            node_xp = mesh["xps"][y, x]

            node_xp["node_info"]["coord"]["x"] = x
            node_xp["node_info"]["coord"]["y"] = y
            node_xp["node_info"]["nodeid"] = xp.get("id", 0)
            node_xp["node_info"]["logicalid"] = xp.get("logical_id", 0)
            node_xp["node_info"]["type"] = CMN_XP_DEVICE_ID
            node_xp["node_info"]["type_str"] = "XP"

            node_xp["dtc_domain"] = xp.get("dtc", 0)
            node_xp["num_device_ports"] = xp.get("n_ports", 0)

            for port in xp.get("ports", [])[:CMN_MAX_PORTS]:
                if "port" not in port or not (0 <= port["port"] < CMN_MAX_PORTS):
                    continue

                po = node_xp["ports"][port["port"]]
                po["type"] = port.get("type", 0)
                po["type_str"] = port.get("type_s", "unknown")[:CMN_TYPE_STR_MAX_LEN]
                po["cal"] = port.get("cal", False)

                devices_list = port.get("devices", [])[:CMN_MAX_CHILDS]
                po["num_devices"] = len(devices_list)

                for d, device in enumerate(devices_list):
                    dev = po["devices"][d]
                    dev["coord"]["x"] = x
                    dev["coord"]["y"] = y
                    dev["nodeid"] = device.get("id", 0)
                    dev["logicalid"] = device.get("logical_id", 0)
                    dev["type"] = device.get("type", 0)
                    dev["type_str"] = device.get("type_s", "unknown")[:CMN_TYPE_STR_MAX_LEN]

    def get_view(
        self,
        mesh_idx: int,
        min_row: int, max_row: int,
        min_col: int, max_col: int
    ):
        """
        Get a view of the mesh array for the visible region (NumPy array slicing).
        Args:
            mesh_idx (int): Mesh index
            min_row (int): Start row (inclusive)
            max_row (int): End row (inclusive)
            min_col (int): Start col (inclusive)
            max_col (int): End col (inclusive)
        Returns:
            NDArray[np.ndarray]: Sliced 2D view of the mesh XP array
        """
        mesh = self.meshes[mesh_idx]
        x_dim = mesh["x_dim"]
        y_dim = mesh["y_dim"]

        min_row = max(0, min_row)
        max_row = min(y_dim, max_row + 1)
        min_col = max(0, min_col)
        max_col = min(x_dim, max_col + 1)

        return mesh["xps"][min_row:max_row, min_col:max_col]
